/*\
//	BMP.c	( Windows 3.0 uncompressed format only )
//
//	based on
 *
 * ppmtobmp.c - Converts from a PPM file to a Microsoft Windows or OS/2
 * .BMP file.
 *
 * Copyright (C) 1992 by David W. Sanderson.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  This software is provided "as is"
 * without express or implied warranty.
 *
 * Revision 1.9  1992/11/24  19:39:33  dws
 *
//
*/

#include	"defs.h"
#include	"BMP.h"
#include	"pislice.h"

#define	MAXCOLORS 256


/* prototypes */
extern int	loadBMPfile( picsliceinfo* );
extern int	drawBMP( picsliceinfo* );
extern void	EndOfDraw( picsliceinfo* );

static int	littlelong( int a );
static short	littleshort( short a );
static int	BMPwritefileheader( FILE *fp, unsigned long bitcount,
									unsigned long x, unsigned long y );
static int	BMPwritergbtable( FILE *fp, int bpp, int colors,
				unsigned short *R, unsigned short *G, unsigned short *B );
static void	BMPwriterow( unsigned char *line, unsigned short *row,
				 unsigned long cx, unsigned short bpp );
static int	BMPwritebits( FILE *fp, unsigned long cx, unsigned long cy,
					unsigned short cBitCount, unsigned short *pixels );
static int	colorstobpp( int colors );
static void	BMPEncode( FILE *fp, int x, int y, unsigned short *pixels,
    int colors, unsigned short *R, unsigned short *G, unsigned short *B );


static int
littlelong( a )
int	a;
{
#ifdef BIGENDIAN
	return ( ( a & 0xff ) << 24 )
		 + ( ( ( a >> 8 ) & 0xff ) << 16 )
		 + ( ( ( a >> 16 ) & 0xff ) << 8 )
		 +   ( ( a >> 24 ) & 0xff );
#else
	return a;
#endif
}

static short
littleshort( a )
short	a;
{
#ifdef BIGENDIAN
	return ( ( a & 0xff ) << 8 )
			+ ( ( a >> 8 ) & 0xff );
#else
	return a;
#endif
}

/*
 * BMP writing
 */

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritefileheader( fp, bitcount, x, y )
FILE	*fp;
unsigned long	bitcount;
unsigned long	x;
unsigned long	y;
{
	BMPHEADER	header;

	header.file.head1= 'B';
	header.file.head2= 'M';

	/* cbSize */
	header.file.cbSize= littlelong( BMPlenfile( bitcount, x, y ) );

	/* xHotSpot */
	header.file.xHotSpot= 0;

	/* yHotSpot */
	header.file.yHotSpot= 0;

	/* offBits */
	header.file.offBits= littlelong( BMPoffbits( bitcount ) );

	/* cbFix */
	header.info.cbFix= littlelong( 40 );

	/* cx */
	header.info.cx= littlelong( x );
	/* cy */
	header.info.cy= littlelong( y );
	/* cPlanes */
	header.info.cPlanes= littleshort( 1 );
	/* cBitCount */
	header.info.cBitCount= littleshort( (short)bitcount );

	/*
	 * We've written 16 bytes so far, need to write 24 more
	 * for the required total of 40.
	*/

	header.info.filler[0]= 0;
	header.info.filler[1]= 0;
	header.info.filler[2]= 0;
	header.info.filler[3]= 0;
	header.info.filler[4]= 0;
	header.info.filler[5]= 0;

	fwrite( &header, sizeof( BMPHEADER ), 1, fp );

	return sizeof( BMPHEADER );
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritergbtable( fp, bpp, colors, R, G, B )
FILE	*fp;
int	bpp;
int	colors;
unsigned short	*R;
unsigned short	*G;
unsigned short	*B;
{
	int	i;
	long	ncolors;
	RGBQUAD	rgbtable[MAXCOLORS];

	for( i= 0; i < colors; i++ ){
		rgbtable[i].b= (unsigned char)(B[i]);
		rgbtable[i].g= (unsigned char)(G[i]);
		rgbtable[i].r= (unsigned char)(R[i]);
		rgbtable[i].filler= 0;
	}

	ncolors= ( 1 << bpp );

	for( ; i < ncolors; i++ ){
		rgbtable[i].b= 0;
		rgbtable[i].g= 0;
		rgbtable[i].r= 0;
		rgbtable[i].filler= 0;
	}

	fwrite( rgbtable, sizeof( RGBQUAD ), ncolors, fp );

	return ncolors * 4;
}

static void
BMPwriterow( line, row, cx, bpp )
unsigned char	*line;
unsigned short	*row;
unsigned long	cx;
unsigned short	bpp;
{
	unsigned int	nbyte= 0;
	unsigned int	x;
	unsigned short	*p= row;

	switch( bpp ){
	case 1:{
		unsigned short	*w= (unsigned short*)line;
		unsigned short	b;
		char	nowbit= 16;

		for( x= 0; x < cx; x++ ){
			if( !nowbit-- ){
				*w++= b;
				nowbit= 16 - 1;
			}
			b*= 2;
			b+= *p++;
			nowbit--;
		}
		if( nowbit ){
			b= b << nowbit;
		}
		*w++= b;
		nbyte= ( cx + 7 ) >> 3;
		break;
	}
	case 4:{
		unsigned short	*w= (unsigned short*)line;
		unsigned short	b;
		char	nowbit= 4;

		for( x= 0; x < cx; x++ ){
			if( !nowbit-- ){
				*w++= b;
				nowbit= 4 - 1;
			}
			b= b << 4;
			b+= *p++;
		}
		if( nowbit ){
			b= b << ( nowbit << 2 );
		}
		*w++= b;
		nbyte= ( cx + 1 ) >> 1;
		break;
	}
	case 8:{
		unsigned char	*w= line;
		for( x= 0; x < cx; x++ ){
			*w++= *p++;
		}
		nbyte= cx;
		break;
	}
	default:{
		unsigned char	*w= line;
		unsigned char	b;
		for( x= 0; x < cx; x++, p++ ){
			b= ( *p & 0x1f ) << 3;
			if( b & 8 ){
				b|= 7;
			}
			*w++= b;
			b= ( ( *p >> 10 ) & 0x1f ) << 3;
			if( b & 8 ){
				b|= 7;
			}
			*w++= b;
			b= ( ( *p >> 5 ) & 0x1f ) << 3;
			if( b & 8 ){
				b|= 7;
			}
			*w++= b;
		}
		nbyte= cx * 3;
		break;
	}
	}

	/*
	 * Make sure we write a multiple of 4 bytes.
	*/
	while( nbyte & 3 ){
		row[nbyte++]= 0;
	}
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritebits( fp, cx, cy, cBitCount, pixels )
FILE	*fp;
unsigned long	cx;
unsigned long	cy;
unsigned short	cBitCount;
unsigned short	*pixels;
{
	long	y;
	int	cols= BMPlenline( cBitCount, cx );
	unsigned char *line= (unsigned char*)MALLOC( cols );

	if( (int)line < 0 )	return -1;

	/*
	 * The picture is stored bottom line first, top line last
	 */

	for( y= cy - 1; y >= 0; y-- ){
		BMPwriterow( line, pixels + y * cx, cx, cBitCount );
		fwrite( line, sizeof( unsigned char ), cols, fp );
	}

	MFREE( line );
	return cols * cy;
}

/*
 * Return the number of bits per pixel required to represent the
 * given number of colors.
 */

static int
colorstobpp(colors)
int	colors;
{
	int	bpp= 24;

	if( !colors ){
		error( "%s: can't have less than one color" );
	}else if( colors <= 2 ){
		bpp= 1;
	}else if( colors <= 16 ){
		bpp= 4;
	}else if( colors <= 256 ){
		bpp= 8;
	}else if( colors >> 24 ){
		error( "%s: too many colors" );
	}

	return bpp;
}

/*
 * Write a BMP file of the given class.
 *
 * Note that we must have 'colors' in order to know exactly how many
 * colors are in the R, G, B, arrays.  Entries beyond those in the
 * arrays are undefined.
 */
static void
BMPEncode( fp, x, y, pixels, colors, R, G, B )
FILE	*fp;
int	x;
int	y;
unsigned short	*pixels;
int	colors;		/* number of valid entries in R,G,B */
unsigned short	*R;
unsigned short	*G;
unsigned short	*B;
{
	int	bpp;	/* bits per pixel */
	unsigned long	nbyte= 0;

	bpp= colorstobpp( colors );

	/*
	 * I have found empirically at least one BMP-displaying program
	 * that can't deal with (for instance) using 3 bits per pixel.
	 * I have seen no programs that can deal with using 3 bits per
	 * pixel.  I have seen programs which can deal with 1, 4, and
	 * 8 bits per pixel.
	 *
	 * Based on this, I adjust actual the number of bits per pixel
	 * as follows.  If anyone knows better, PLEASE tell me!
	*/
	switch( bpp ){
/*	case 1:
		MaxColorGuaranteed= 2;
		break;/*
	case 2:
	case 3:
		bpp= 4;
	case 4:
/*		MaxColorGuaranteed= 16;*/
		break;
	case 5:
	case 6:
	case 7:
		bpp= 8;
/*		MaxColorGuaranteed= 65536;*/
		break;
	}

	nbyte += BMPwritefileheader( fp, bpp, x, y );
	if( bpp != 24 ){
		nbyte += BMPwritergbtable( fp, bpp, colors, R, G, B );
	}

	if( nbyte != ( BMPlenfileheader
				 + BMPleninfoheader
				 + BMPlenrgbtable( bpp ) ) ){
		error( "%s: abnormal data" );
	}

	nbyte += BMPwritebits( fp, x, y, bpp, pixels );
		if( nbyte != BMPlenfile( bpp, x, y ) ){
			error( "%s: BMPEncode failed" );
		}
}


void
BMPSave()
{
	int	i, HalfSize;
	short	*ctable;

	int MaxColor= 0;

	if( (int)( ctable= (short*)MALLOC( 32768 * 2 ) ) < 0 ){
		error( "%s: memory exhausted" );
	}
	for( i= 0; i < 32768; i++ ){
		ctable[i]= -1;
	}

	HalfSize= Size / 2;
	for( i= 0; i < HalfSize; i++ ){
		unsigned short c= Screen[i];
		if( ctable[c] < 0 ){
			Palette[MaxColor]= c;
			ctable[c]= MaxColor;
			if( ++MaxColor > MAXCOLORS ){
				break;
			}
		}
	}

	if( MaxColor <= MAXCOLORS ){
		unsigned short	*p= Screen;
		unsigned short	*w= Screen;

		for( i= 0; i < MaxColor; i++ ){
			Green[i]= ( ( Palette[i] >> 10 ) & 0x1f ) << 3;
			if( Green[i] & 8 ){
				Green[i]|= 7;
			}
			Red[i]= ( ( Palette[i] >> 5 ) & 0x1f ) << 3;
			if( Red[i] & 8 ){
				Red[i]|= 7;
			}
			Blue[i]= ( Palette[i] & 0x1f ) << 3;
			if( Blue[i] & 8 ){
				Blue[i]|= 7;
			}
		}
		for( i= 0; i < HalfSize; i++ ){
			*w++= ctable[*p++];
		}
	}
	MFREE( ctable );

    if( fp == NULL && ( fp= fopen( File, "wb" )) == NULL ){
		error("%s: Cannot create");
	}

	/* All set, let's do it. */
	BMPEncode( fp, Width, Height, Screen, MaxColor,
				Red, Green, Blue);

	if( fp != stdout ){
		fclose (fp);
	}
}

void
BMPLoad()
{
	picsliceinfo	psi;
	int	ret;

	if( fp ){
		psi.filehandle= 0;
	}else if( (psi.filehandle= OPEN( File, 0 )) < 0 ){
		error( "%s: No such file" );
	}
	if( ( ret= loadBMPfile( &psi ) ) != PICSLICE_NORMAL ){
		if( ret == PICSLICE_NOMEMORY ){
			error ("%s: Insufficient memory");
		}else {
			error( "%s: Is not bmp format" );
		}
	}
	Width= psi.width;
	Height= psi.height;
	MaxColorGuaranteed= psi.realwidth;

	/*	AllocateScreen()	*/
    if( Screen )	MFREE( Screen );
    if( (int)( Screen= (int16*)MALLOC( ( Size= Width * Height * 2 ) + 256 ) ) < 0 ){
		EndOfDraw( &psi );
		error ("%s: Insufficient memory");
	}

	(unsigned short*)psi.chainbuf= Screen;
	if( ( ret= drawBMP( &psi ) ) != PICSLICE_TERMINATE ){
		error( "%s: load error" );
	}
	RightShift();
}
