/*
//	Pic.h -- pic$@%U%)!<%^%C%H2hA|$r07$&(J
//
//		created    in 3/1/1993
//		revised    in 2/13/1994
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	"defs.h"
#include	"Pic.h"
#include	"PICSLICE.H"

#define	SetMark(x,y)	(Mark[(x)/8 + (y) * PicRadix] |= BitField[(x)%8])
#define	ResetMark(x,y)	(Mark[(x)/8 + (y) * PicRadix] &= ~BitField[(x)%8])
#define	FetchMark(x,y)	(Mark[(x)/8 + (y) * PicRadix] & BitField[(x)%8])

struct	{
    int16 	Color;
    int 	Next;
    int 	Previous;
}	LRUCache[128];

unsigned char	*BitBufferPointer;
int32	PicRadix;
unsigned char	*Mark;
uint32 	PicWord;
int 	NoExpand;
int 	ColorPointer;

extern int	loadPICfile( picsliceinfo* );
extern int	drawPIC( picsliceinfo* );
extern void	EndOfDraw( picsliceinfo* );
void	ColorCachInitialize();
int16 	NewColor( int16 c );
void	WritePicHeader();
void	DifferentPointMark();
void	CompressPic();
void	CompressChain( int x, int y, int c );
void	WriteLength( uint32 Length );
void	WriteColor( int16 c );
int 	SearchColor( int16 c );
void	SetColor( int i );

void	PicLoad()
{
	picsliceinfo	psi;
	int	ret;
    int 	OldWidth, x, y;

	if( fp ){
		psi.filehandle= 0;
	}else if( (psi.filehandle= OPEN( File, 0 )) < 0 ){
		error( "%s: No such file" );
	}
	if( ( ret= loadPICfile( &psi ) ) != PICSLICE_NORMAL ){
		if( ret == PICSLICE_NOMEMORY ){
			error ("%s: Insufficient memory");
		}else {
			error( "%s: Is not pic format" );
		}
	}
	Width= psi.width;
	Height= psi.height;

	/*	AllocateScreen()	*/
    if( Screen )	MFREE( Screen );
	if( !NoExpand ){
		OldWidth = Width;
		Width = Width * 4 / 3;
	}
	Size= Width * Height * 2;
    if( (int)( Screen= (int16*)MALLOC( Size ) ) < 0 ){
		EndOfDraw( &psi );
		error ("%s: Insufficient memory");
	}

	(unsigned short*)psi.chainbuf= Screen;
	if( ( ret= drawPIC( &psi ) ) != PICSLICE_TERMINATE ){
		error( "%s: load error" );
	}
	RightShift();

    if ( !NoExpand ) {
	for ( y = Height - 1; y >= 0; y-- ) {
	    for ( x = Width - 1; x >= 0; x-- ) {
		Screen[x + y * Width] = Screen[x * 3 / 4 + y * OldWidth];
	    }
	}
    }
    MaxColorGuaranteed = 32768;
}

void	ColorCachInitialize()
{
    int 	i;

    for ( i = 0; i < 128; i++ ) {
	LRUCache[i].Color = 0;
	LRUCache[i].Previous = i + 1;
	LRUCache[i].Next = i - 1;
    }
    LRUCache[127].Previous = 0;
    LRUCache[0].Next = 127;
    ColorPointer = 0;
}

void	PicMarkClear()
{
    int 	i;

    for ( i = 0; i < PicRadix * Height; i++ )
	Mark[i] = 0;
}

int16 	NewColor( int16 c )
{
    ColorPointer = LRUCache[ColorPointer].Previous;
    LRUCache[ColorPointer].Color = c;
    return c;
}


void	PicSave()
{
    int 	OldWidth, x, y;

    if ( fp == NULL && (fp = fopen( File, "wb" )) == NULL )
	error ("%s: Cannot create");
    AllocateBufferForWriting();
    BitBufferLength = BitBufferSize;
    BitLength = 8;
    BitBufferPointer = BitBuffer;
    ColorCachInitialize();
    if ( !NoExpand ) {
	OldWidth = Width;
	Width = Width * 3 / 4;
	for ( y = 0; y < Height; y++ ) {
	    for ( x = 0; x < OldWidth; x++ ) {
		Screen[x * 3 / 4 + y * Width] = Screen[x + y * OldWidth];
	    }
	}
    }
    WritePicHeader();
    PicRadix = (Width + 7) / 8;
    if ( (int)(Mark = (unsigned char*)MALLOC (PicRadix * Height)) < 0 )
	error ("%s: Insufficient memory");
    PicMarkClear();
    DifferentPointMark();
    CompressPic();
    PicBufferWriteFlush();
    MFREE (Mark);
    MFREE (BitBuffer);
    if ( fp != stdout )
	fclose (fp);
}

void	WritePicHeader()
{
    fputc( 'P', fp );
    fputc( 'I', fp );
    fputc( 'C', fp );

    if( NoExpand ){
	    fputc( '/', fp );	/* square pic header( mmapic ) */
	    fputc( 'M', fp );
	    fputc( 'M', fp );
	    fputc( '/', fp );
	    fputc( 'X', fp );
	    fputc( 'S', fp );
	    fputc( 'S', fp );
	    fputc( '/', fp );
	    fputc( ':', fp );
    }

    fputc( 26, fp );
    fputc( 0, fp );
    fputc( 0, fp ); fputc( 0, fp );
    fputc( 0, fp ); fputc( 15, fp );
    fputc( Width >> 8, fp ); fputc( Width & 0xff, fp );
    fputc( Height >> 8, fp ); fputc( Height & 0xff, fp );
}

void	DifferentPointMark()
{
    int 	x, y, c, a;

    c = 0;
    for ( y = 0; y < Height; y++ ) {
	for ( x = 0; x < Width; x++ ) {
	    a = Screen[x + y * Width];
	    if ( a != c ) {
		c = a;
		SetMark( x, y );
	    }
	    else
		ResetMark( x, y );
	}
    }
}

void	CompressPic()
{
    uint32	Length;
    int 	x, y;
    int16	a;

    Length = 0;
    for ( y = 0; y < Height; y++ ) {
	for ( x = 0; x < Width; x++ ) {
	    Length++;
	    if ( FetchMark( x, y ) != 0 ) {
		a = Screen[x + y * Width];
		WriteLength (Length);
		WriteColor (a);
		CompressChain( x, y, a );
		ResetMark( x, y );
		Length = 0;
	    }
	}
    }
    WriteLength (Length + 1);
}

void	CompressChain( int x, int y, int c )
{
    int 	yy, d, f;

    f = 0;
    for ( yy = y + 1; yy < Height; yy++ ) {
	if ( Screen[x + yy * Width] == c )
	    d = 2;
	else if ( --x >= 0     && Screen[x + yy * Width] == c )
	    d = 1;
	else if ( (x+=2) < Width && Screen[x + yy * Width] == c )
	    d = 3;
	else if ( (x-=3) >= 0  && Screen[x + yy * Width] == c )
	    d = 4;
	else if ( (x+=4) < Width && Screen[x + yy * Width] == c )
	    d = 5;
	else
	    break;
	ResetMark( x, yy );
	if ( f == 0 )
	    PicBitWrite1( (uint32)1 );
	if ( d == 4 )
	    PicBitWrite( 4, (uint32)2 );
	if ( d == 5 )
	    PicBitWrite( 4, (uint32)3 );
	if ( d <= 3 )
	    PicBitWrite( 2, (uint32)d );
	f = 1;
    }
    if ( f == 0 ) {
	PicBitWrite1( (uint32)0 );
    }
    else {
	PicBitWrite( 3, (uint32)0 );
    }
}

void	WriteLength( uint32 Length )
{
    int 	a;
    uint32	b;

    a = 1;
    b = 4;
    while ( Length > b - 2 ) {
	a++;
	b = b * 2;
    }
    PicBitWrite( a, 0xfffffffe );
    PicBitWrite( a, (uint32)(Length + 1 - b / 2) );
}

void	WriteColor( int16 c )
{
    int 	a;

    if ( (a = SearchColor (c)) != -1 ) {
	PicBitWrite( 8, (uint32)(a + 128) );
    }
    else {
	PicBitWrite1( 0 );
	PicBitWrite( 15, (uint32)c );
    }
}

int 	SearchColor( int16 c )
{
    int 	i;

    for ( i = 0; i < 128; i++ ) {
	if ( LRUCache[i].Color == c )
	    break;
    }
    if ( i == 128 ) {
	NewColor (c);
	i = -1;
    }
    else
	SetColor( i );
    return (i);
}

void	SetColor( int i )
{
    if ( ColorPointer != i ) {
	LRUCache[LRUCache[i].Previous].Next = LRUCache[i].Next;
	LRUCache[LRUCache[i].Next].Previous = LRUCache[i].Previous;
	LRUCache[LRUCache[ColorPointer].Previous].Next = i;
	LRUCache[i].Previous = LRUCache[ColorPointer].Previous;
	LRUCache[ColorPointer].Previous = i;
	LRUCache[i].Next = ColorPointer;
	ColorPointer = i;
    }
}

void	PicWriteBufferNext()
{
    if ( --BitBufferLength == 0 ) {
	if ( fwrite( BitBuffer, 1, BitBufferSize, fp ) != BitBufferSize )
	    error ("%s: Error in writing");
	BitBufferPointer = BitBuffer;
	BitBufferLength = BitBufferSize;
    }
    else
	++BitBufferPointer;
    BitLength = 8;
}

void	PicBufferWriteFlush()
{
    if ( BitBufferLength > 0 ) {
	*BitBufferPointer <<= BitLength;
	PicWriteBufferNext();
    }
    BitBufferLength = BitBufferSize - BitBufferLength;
    if ( BitBufferLength > 0 ) {
	if ( fwrite( BitBuffer, 1, BitBufferLength, fp ) != BitBufferLength )
	    error ("%s: Error in writing");
    }
}

void	PicBufferNext()
{
    if ( BitBufferLength == 0 ) {
	if ( (BitBufferLength = fread( BitBuffer, 1, BitBufferSize, fp )) == 0 )
	    error ("%s: Unexpected EOF");
	BitBufferPointer = BitBuffer;
    }
    else
	BitBufferPointer++;
    BitBufferLength--;
    BitLength = 8;
}
